/* SPDX-License-Identifier: GPL-2.0-or-later */
/* Copyright (C) 2019 Western Digital Corporation or its affiliates */

#ifndef UFS_ERR_HIST_H_
#define UFS_ERR_HIST_H_
#include "options.h"

void err_hist_help(char *tool_name);
int do_err_hist(struct tool_options *opt);
#endif /*UFS_ERR_HIST_H_*/
