/* SPDX-License-Identifier: GPL-2.0-or-later */
/* Copyright (C) 2019 Western Digital Corporation or its affiliates */

#ifndef UFS_VENDOR_H_
#define UFS_VENDOR_H_

void vendor_help(char *tool_name);
int do_vendor(struct tool_options *opt);

#endif /*UFS_VENDOR_H_*/
